import numpy as np
import teneva
from time import perf_counter as tpc


from .fpcross import FPCross
from .utils import _renormalize_new


class FPCrossOUPDiag(FPCross):
    def _apply_exact_oup(self, G, X, rhs, t, cf=1, mu=0.):
        X0 = X
        X0, X1 = np.meshgrid(X0, X, indexing='xy')
        c = (1 - np.exp(-2. * rhs * t)) / rhs
        M = -((mu - X1 + (X0 - mu)*np.exp(-rhs*t))**2) / c
        M = cf*np.exp(M) / np.sqrt(np.pi * c)
        return M @ G

    def _step(self):
        tc = tpc()
        mu, a, b, n = self.eq.coef_mu, self.eq.a, self.eq.b, self.eq.n
        args = zip(self.Y0, self.X, np.diag(self.eq.coef_rhs), a, b, n)
        self.Y = [self._apply_exact_oup(G, X, rhs, self.t, (b-a)/n)
            for k, (G, X, rhs, a, b, n) in enumerate(args)]

        self.tc += tpc() - tc

        self._step_hist()

    def _step_init(self):
        tc = tpc()
        self.X = [np.linspace(a, b, n)
            for a, b, n in zip(self.eq.a, self.eq.b, self.eq.n)]
        self.Y0 = self.eq.build_r0()
        self.Y0 = _renormalize_new(self.Y0, self.eq.a, self.eq.b, self.eq.n)
        self.tc += tpc() - tc



"""Apply the Fokker-Planck solver for different model problems.

Run it from the root of the project as "clear && python demo/check.py". The
output should look like this (see also the folder "demo/result" with plots):
"
>>>>>  DIF-3D
Solve: 100step [00:06, 15.47step/s,  | t=2.50 | r=1.0 | e_t=1.50e-02]
-----  Solved > Time :   7.0189 sec

>>>>>  OUP-3D
Solve: 100step [00:08, 11.91step/s,  | t=5.00 | r=4.0 | e_s=1.57e-03]
-----  Solved > Time :   8.7419 sec

>>>>>  OUP-5D
Solve: 100step [00:39,  2.56step/s,  | t=5.00 | r=4.5 | e_s=1.95e-03]
-----  Solved > Time :  39.4404 sec

>>>>>  DIF-3D (Numpy format)
Solve: 100step [00:58,  1.72step/s,  | t=2.50 | e_t=1.50e-02]
-----  Solved > Time :  58.3198 sec

>>>>>  OUP-1D (Numpy format)
Solve: 100step [00:00, 644.19step/s,  | t=5.00 | e_s=3.59e-04 | e_t=3.59e-04]
-----  Solved > Time :   0.4358 sec

>>>>>  OUP-2D (Numpy format)
Solve: 100step [00:02, 36.88step/s,  | t=5.00 | e_s=7.07e-04]
-----  Solved > Time :   3.0187 sec

>>>>>  OUP-3D (Numpy format)
Solve: 100step [03:44,  2.25s/step,  | t=5.00 | e_s=1.98e-03]
-----  Solved > Time : 225.1532 sec

>>>>>  DUM-3D
Solve: 100step [01:20,  1.24step/s,  | t=10.00 | r=7.9 | e_eta=1.05e-03 | e_psi=3.72e-04]
-----  Solved > Time :  80.9169 sec
".

"""
import numpy as np
import os
from time import perf_counter as tpc


from fpcross import EquationDif
from fpcross import EquationDum
from fpcross import EquationOUP
from fpcross import FPCross


def run():
    print(f'>>>>>  DIF-3D')
    t = tpc()
    run_dif_3d()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')

    print(f'>>>>>  OUP-3D')
    t = tpc()
    run_oup_3d()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')

    print(f'>>>>>  OUP-5D')
    t = tpc()
    run_oup_5d()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')


def run_np():
    print(f'>>>>>  DIF-3D (Numpy format)')
    t = tpc()
    run_dif_3d_np()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')

    print(f'>>>>>  OUP-1D (Numpy format)')
    t = tpc()
    run_oup_1d_np()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')

    print(f'>>>>>  OUP-2D (Numpy format)')
    t = tpc()
    run_oup_2d_np()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')

    print(f'>>>>>  OUP-3D (Numpy format)')
    t = tpc()
    run_oup_3d_np()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')


def run_spec():
    print(f'>>>>>  DUM-3D')
    t = tpc()
    run_dum_3d()
    print(f'-----  Solved > Time : {tpc()-t:-8.4f} sec\n')


def run_dif_3d():
    eq = EquationDif(d=3)
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/dif_3d')


def run_dif_3d_np():
    eq = EquationDif(d=3, is_full=True)
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/dif_3d_np')


def run_dum_3d():
    eq = EquationDum(d=3)
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/dum_3d', is_spec=True)


def run_oup_1d_np():
    eq = EquationOUP(d=1, is_full=True)
    eq.set_coef_rhs(np.array([
        [1.]
    ]))
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/oup_1d_np')


def run_oup_2d_np():
    eq = EquationOUP(d=2, is_full=True)
    eq.set_coef_rhs(np.array([
        [1.0, 0.5],
        [0.0, 1.0],
    ]))
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/oup_2d_np')


def run_oup_3d():
    eq = EquationOUP(d=3)
    eq.set_coef_rhs(np.array([
        [1.5, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.5, 0.3, 1.0],
    ]))
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/oup_3d')


def run_oup_3d_np():
    eq = EquationOUP(d=3, is_full=True)
    eq.set_coef_rhs(np.array([
        [1.5, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.5, 0.3, 1.0],
    ]))
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/oup_3d_np')


def run_oup_5d():
    eq = EquationOUP(d=5)
    eq.set_coef_rhs(np.array([
        [1.5, 0.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 1.0, 0.0],
        [0.5, 0.3, 0.2, 0.0, 1.0],
    ]))
    eq.init()

    fpc = FPCross(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/oup_5d')


if __name__ == '__main__':
    os.makedirs('./demo/result', exist_ok=True)

    run()
    run_np()
    run_spec()



import numpy as np
import teneva


class Equation:
    def __init__(self, d, e, is_full=False, name='Equation'):
        """Class that represents the Fokker-Planck equation.

        Args:
            d (int): number of dimensions.
            e (float): desired accuracy of approximation. It will be used for
                all truncations of the TT-tensors. Note that this parameter is
                not used if "is_full" flag is set.
            is_full (bool): if flag is set, then full (numpy) format will be
                used instead of the TT-format. Note that the full format may be
                used only for small dimension numbers ("d").
            name (str): optional display name for the equation.

        """
        self.d = int(d)
        self.e = float(e)
        self.is_full = bool(is_full)
        self.name = str(name)

        self.set_coef()
        self.set_coef_pdf()
        self.set_cross_opts()
        self.set_grid()
        self.set_grid_time()

        self.Ys = None
        self.Yt = None

    def build(self, func, Y0=None, e=None):
        """Build the function on the Chebyshev grid related to equation.

        Args:
            func (function): function f(X) for interpolation, where X should be
                2D np.ndarray of the shape [samples, dimensions]. The function
                should return 1D np.ndarray of the length equals to "samples".
            Y0 (list): TT-tensor, which is the initial approximation for the
                TT-cross algorithm. If it is not set, then random TT-tensor of
                the TT-ranks equals "self.cross_r" will be used.
            e (float): convergence criterion for the TT-cross algorithm. If
                between iterations the relative rate of solution change is less
                than this value, then the operation of the algorithm will be
                interrupted. If is not set, then "self.e" option will be used.

        Returns:
            list: TT-Tensor with function values on the Chebyshev grid. If full
            format is used (see flag "is_full"), then the d-dimensional
            np.ndarray with function values on the Chebyshev grid will be
            returned.

        Note:
            You can use the "set_cross_opts" function to refine the parameters
            of the TT-cross method.

        """
        self.cross_info = {}
        self.cross_cache = {} if self.cross_with_cache else None

        if Y0 is None:
            Y0 = teneva.rand(self.n, self.cross_r)

        if self.is_full:
            Y = teneva.cheb_bld_full(func, self.a, self.b, self.n)
        else:
            e = e or self.e
            Y = teneva.cheb_bld(func, self.a, self.b, self.n, eps=e, Y0=Y0,
                m=self.cross_m,
                e=e,
                nswp=self.cross_nswp,
                tau=self.cross_tau,
                dr_min=self.cross_dr_min,
                dr_max=self.cross_dr_max,
                tau0=self.cross_tau0,
                k0=self.cross_k0,
                info=self.cross_info,
                cache=self.cross_cache)

        return Y

    def build_r0(self):
        """Build the initial PDF r(x, 0) on the Chebyshev grid.

        Returns:
            list: TT-Tensor with function values on the Chebyshev grid. If full
            format is used (see flag "is_full"), then the d-dimensional
            np.ndarray with function values on the Chebyshev grid will be
            returned.

        Note:
            The function specified in the "r0" method is used.

        """
        func = self.r0
        self.Y0 = self.build(func)
        return self.Y0

    def build_rs(self):
        """Build the stationary PDF r0(x) on the Chebyshev grid.

        Returns:
            list: TT-Tensor with function values on the Chebyshev grid. If full
            format is used (see flag "is_full"), then the d-dimensional
            np.ndarray with function values on the Chebyshev grid will be
            returned. If flag "with_rs" is not set, then it returns None.

        """
        if not self.with_rs:
            self.Ys = None
            return self.Ys

        func = self.rs
        self.Ys = self.build(func, e=self.e/100)
        return self.Ys

    def build_rt(self, t):
        """Build the time-dependent PDF r(x, t) on the Chebyshev grid.

        Args:
            t (float): the current value of time.

        Returns:
            list: TT-Tensor with function values on the Chebyshev grid. If full
            format is used (see flag "is_full"), then the d-dimensional
            np.ndarray with function values on the Chebyshev grid will be
            returned. If flag "with_rt" is not set, then it returns None.

        """
        if not self.with_rt:
            self.Yt = None
            return self.Yt

        func = lambda X: self.rt(X, t)
        self.Yt = self.build(func, e=self.e/100)
        return self.Yt

    def callback(self, solver):
        """Callback function, which is called by solver after each time step.

        Args:
            solver (FPCross): Fokker-Planck solver class instance.

        """
        return

    def f(self, X, t):
        """The function computes the rhs of equation.

        Args:
            X (np.ndarray): the set of spatial points. It is np.ndarray of the
                shape [samples, dimensions].
            t (float): the current value of time.

        Returns:
            np.ndarray: the values of the vector-values rhs for all samples (it
            is 2D array of the shape [samples, dimensions]).

        """
        return np.zeros(X.shape)

    def f1(self, X, t):
        """The function computes the derivatives for rhs of equation.

        Args:
            X (np.ndarray): the set of spatial points. It is np.ndarray of the
                shape [samples, dimensions].
            t (float): the current value of time.

        Returns:
            np.ndarray: the values of the "diagonal" partial derivatives for
            the rhs ( d f_1 / d x_1, d f_2 / d x_2, ..., d f_d / d x_d) for all
            samples (it is 2D array of the shape [samples, dimensions]).

        """
        return np.zeros(X.shape)

    def init(self):
        """The function is used to initialize the equation parameters.

        Run it before the solution by FPCross solver but after the setting all
        equation parameters ("set_coef", "set_grid", etc.).

        """
        self.with_rs = False
        self.with_rt = False

    def r0(self, X):
        """The function computes initial PDF r(x, 0).

        Args:
            X (np.ndarray): the set of spatial points. It is np.ndarray of the
                shape [samples, dimensions].

        Returns:
            np.ndarray: the values of the initial PDF r(x, 0) for all samples
            (it is 1D array of the shape [samples]).

        Note:
            This function can be changed in a child class. Here we use the
            Gaussian distribution with parameter "coef_pdf", which is defined
            in "set_coef_pdf" function.

        """
        a = 2. * self.coef_pdf
        r = np.exp(-np.sum(X*X, axis=1) / a)
        r /= (np.pi * a)**(self.d/2)
        return r

    def rs(self, X):
        """The function computes stationary PDF r0(x).

        Args:
            X (np.ndarray): the set of spatial points. It is np.ndarray of the
                shape [samples, d].

        Returns:
            np.ndarray: the values of the stationary PDF r0(x) for all samples
            (it is 1D array of the shape [samples]).

        Note:
            This function can be defined in a child class. In this case, the
            flag "self.with_rs" must also be set manually to True value (e.g.,
            inside the "init" function). In this case, the Fokker-Planck solver
            will automatically check the accuracy of the received solution.

        """
        return

    def rt(self, X, t):
        """The function computes time-dependent PDF r(x, t).

        Args:
            X (np.ndarray): the set of spatial points. It is np.ndarray of the
                shape [samples, dimensions].
            t (float): the current value of time.

        Returns:
            np.ndarray: the values of the time-dependent PDF r(x, t) for all
            samples (it is 1D array of the shape [samples]).

        Note:
            This function can be defined in a child class. In this case, the
            flag "self.with_rt" must also be set manually to True value (e.g.,
            inside the "init" function). In this case, the Fokker-Planck solver
            will automatically check the accuracy of the received solution.

        """
        return

    def set_coef(self, coef=0.5):
        """Set diffusion coefficient.

        Args:
            coef (float): scalar diffusion coefficient.

        """
        self.coef = coef

    def set_coef_pdf(self, coef_pdf=1.):
        """Set coefficient for initial PDF Gaussian distribution.

        Args:
            coef_pdf (float): coefficient for initial PDF.

        """
        self.coef_pdf = coef_pdf

    def set_cross_opts(self, m=None, e=None, nswp=10, tau=1.1, dr_min=1, dr_max=1, tau0=1.05, k0=100, with_cache=True, r=1):
        """Set parameters for the TT-cross method.

        See "https://teneva.readthedocs.io/code/core/cross.html" with a
        detailed description of the parameters. In addition to this parameters,
        function also accepts "with_cache" (if flag is True, then cache for
        TT-cross will be used) and "r" (TT-rank for initial approximation).

        """
        self.cross_m = m
        self.cross_e = e
        self.cross_nswp = nswp
        self.cross_tau = tau
        self.cross_dr_min = dr_min
        self.cross_dr_max = dr_max
        self.cross_tau0 = tau0
        self.cross_k0 = k0
        self.cross_with_cache = with_cache
        self.cross_r = r

    def set_grid(self, n=10, a=-1., b=+1.):
        """Set spatial grid parameters for the discretization.

        Args:
            n (int, float, list, np.ndarray): grid size for each dimension
                (list or np.ndarray of length "d"). It may be also int/float,
                then the size for each dimension will be the same.
            a (float, list, np.ndarray): lower bounds for each dimension (list
                or np.ndarray of length "d"). It may be also float, then the
                lower bounds for each dimension will be the same.
            b (float, list, np.ndarray): upper bounds for each dimension (list
                or np.ndarray of length "d"). It may be also float, then the
                upper bounds for each dimension will be the same.

        Note:
            Only "unfirom" grids are supported now (the size for each dimension
            should be the same), i.e., n / a / b should be int/float.

        """
        if isinstance(n, (int, float)):
            n = np.ones(self.d, dtype=int) * int(n)
        else:
            raise ValueError('Only "unfirom" grids are supported now')
        self.n = np.asanyarray(n, dtype=int)
        if len(self.n) != self.d:
            raise ValueError('Invalid length of n')

        if isinstance(a, (int, float)):
            a = [a] * self.d
        else:
            raise ValueError('Only "unfirom" grids are supported now')
        self.a = np.asanyarray(a, dtype=float)
        if len(self.a) != self.d:
            raise ValueError('Invalid length of a')

        if isinstance(b, (int, float)):
            b = [b] * self.d
        else:
            raise ValueError('Only "unfirom" grids are supported now')
        self.b = np.asanyarray(b, dtype=float)
        if len(self.b) != self.d:
            raise ValueError('Invalid length of b')

    def set_grid_time(self, m=100, t=1.):
        """Set time grid parameters for the discretization.

        Args:
            m (int, float): grid size (i.e., the number of time steps).
            t (float): upper bound for time grid.

        """
        self.m = int(m)
        self.t = t
        self.h = self.t / self.m



import numpy as np
from scipy.linalg import solve_lyapunov


from ..equation import Equation


class EquationOUP(Equation):
    """Ornstein-Uhlenbeck process, a special case of the FPE."""

    def __init__(self, d=3, e=1.E-4, is_full=False, name='OUP'):
        super().__init__(d, e, is_full, name)

        self.set_coef_rhs()
        self.set_grid(n=30, a=-5., b=+5.)
        self.set_grid_time(m=100, t=5.)

    def f(self, X, t):
        return -X @ self.coef_rhs.T

    def f1(self, X, t):
        shape = X.shape[:-1] + (1, )  # Tiling directions for a row.
        grads = np.tile(-np.diag(self.coef_rhs), shape)
        return grads

    def init(self):
        self.with_rs = True
        self.with_rt = True if self.d == 1 else False

        self.W = solve_lyapunov(self.coef_rhs, 2. * self.coef * np.eye(self.d))
        self.Wi = np.linalg.inv(self.W)
        self.Wd = np.linalg.det(self.W)

    def rs(self, X):
        r = np.exp(-0.5 * np.diag(X @ self.Wi @ X.T))
        r /= np.sqrt(2**self.d * np.pi**self.d * self.Wd)
        return r

    def rt(self, X, t):
        if self.d != 1:
            raise ValueError('It is valid only for 1D case')

        K = self.coef_rhs[0, 0]
        S = (1. - np.exp(-2. * K * t)) / 2. / K
        S += self.coef * np.exp(-2. * K * t)
        a = 2. * S
        r = np.exp(-X * X / a) / np.sqrt(np.pi * a)
        return r.reshape(-1)

    def set_coef_rhs(self, coef_rhs=None):
        """Set coefficient matrix for the rhs.

        Args:
            coef_rhs (np.ndarray): coefficient matrix for the rhs (rhs = -1 *
                coef_rhs * x). It is np.ndarray of the shape [d, d].

        """
        if coef_rhs is None:
            self.coef_rhs = np.eye(self.d)
        else:
            self.coef_rhs = coef_rhs



import numpy as np
import teneva


def _renormalize_new(Y, a, b, n):
    a, b, n = teneva.grid_prep_opts(a, b, n)

    c = np.array([1.])
    for ci, a, b, n in zip(Y, a, b, n):
        c = c @ ci.sum(axis=1)
        c *= (b - a) / n

    s = c.item()
    return teneva.mul(1./s, Y)



import numpy as np
import teneva


from .equation_oup import EquationOUP
from ..utils import _renormalize_new


class EquationOUPUniv(EquationOUP):
    def __init__(self, *args, **kwargs):
        self.set_coef_mu(kwargs.pop('mu', 0))

        super().__init__(*args, **kwargs)

    def build(self, func, Y0=None, e=None):
        self.cross_info = {}
        self.cross_cache = {} if self.cross_with_cache else None

        if Y0 is None:
            Y0 = teneva.tensor_rand(self.n, self.cross_r)

        e = e or self.e
        f = lambda I: func(teneva.ind_to_poi(I, self.a, self.b, self.n, 'uni'))
        Y = teneva.cross(f, Y0, self.cross_m, e, self.cross_nswp,
            self.cross_tau, self.cross_dr_min, self.cross_dr_max,
            self.cross_tau0, self.cross_k0, self.cross_info, self.cross_cache)

        return teneva.truncate(Y, e)

    def build_rs_(self):
        # TODO: This function is completely identical to the function from the
        # base class. Why we need it?

        if not self.with_rs:
            self.Ys = None
            return self.Ys

        func = self.rs
        self.cross_nswp = 100
        self.Ys = self.build(func, e=self.e/100)
        self.Ys = _renormalize_new(self.Ys, self.a, self.b, self.n)
        return self.Ys

    def rs_(self, X):
        # TODO: What does it mean?
        self.log_M = (X - self.coef_mu) @ self.Wi @ (X.T - self.coef_mu)
        self.log_M =  -0.5 * np.diag(self.log_M)
        r = np.exp(self.log_M)
        Det = np.sqrt(2**self.d * np.pi**self.d * self.Wd)
        return r / Det

    def set_coef_mu(self, mu=0.):
        self.coef_mu = mu



import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator


COLOR1 = '#8b1d1d'
COLOR2 = '#5f91ac'


def plot(res, fpath=None):
    t = [i*res.eq.h for i in range(1, res.eq.m + 1)]

    if res.is_full:
        fig, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(8, 8))
    else:
        fig, (ax1, ax2) = plt.subplots(nrows=1, ncols=2, figsize=(16, 8))
        plt.subplots_adjust(wspace=0.2)

    ax1.set_title('Relative error')
    ax1.set_xlabel('Time')
    if len(res.es_list) == len(t):
        ax1.plot(t, res.es_list, label='vs stationary',
            linestyle='-', linewidth=2, color=COLOR1,
            marker='o', markersize=7, markerfacecolor=COLOR1,
            markeredgewidth=1, markeredgecolor=COLOR1)
    if len(res.et_list) == len(t):
        ax1.plot(t, res.et_list, label='vs analytic',
            linestyle='-.', linewidth=2, color=COLOR2,
            marker='o', markersize=0, markerfacecolor=COLOR2,
            markeredgewidth=0, markeredgecolor=COLOR2)
    prep(ax1)

    if not res.is_full:
        ax2.set_title('TT-rank')
        ax2.set_xlabel('Time')
        ax2.plot(t, res.r_list,
            linestyle='-', linewidth=2, color=COLOR1,
            marker='o', markersize=7, markerfacecolor=COLOR1,
            markeredgewidth=1, markeredgecolor=COLOR1)
        prep(ax2, with_leg=False, is_log=False)

    if fpath:
        plt.savefig(fpath, bbox_inches='tight')
    else:
        plt.show()


def plot_spec(res, fpath=None):
    t = [i*res.eq.h for i in range(1, res.eq.m + 1)]

    fig, (ax1, ax2) = plt.subplots(nrows=1, ncols=2, figsize=(16, 8))
    plt.subplots_adjust(wspace=0.2)

    with_psi = len(res.eq.psi_list) == len(t)
    with_eta = len(res.eq.eta_list) == len(t)

    ax1.set_title('Computation results')
    ax1.set_xlabel('Time')
    if with_psi:
        ax1.plot(t, res.eq.psi_list, label='Value of $\psi$',
            linestyle='-', linewidth=2, color='#5f91ac',
            marker='o', markersize=7, markerfacecolor='#5f91ac',
            markeredgewidth=1, markeredgecolor='#5f91ac')
    if with_eta:
        ax1.plot(t, res.eq.eta_list, label='Value of $\eta$',
            linestyle='-', linewidth=2, color='#8b1d1d',
            marker='o', markersize=7, markerfacecolor='#8b1d1d',
            markeredgewidth=1, markeredgecolor='#8b1d1d')
    prep(ax1, with_leg=(with_psi or with_eta), is_log=False)

    ax2.set_title('TT-rank')
    ax2.set_xlabel('Time')
    ax2.plot(t, res.r_list,
        linestyle='-', linewidth=2, color='#8b1d1d',
        marker='o', markersize=7, markerfacecolor='#8b1d1d',
        markeredgewidth=1, markeredgecolor='#8b1d1d')
    prep(ax2, with_leg=False, is_log=False)

    if fpath:
        plt.savefig(fpath, bbox_inches='tight')
    else:
        plt.show()


def prep(ax, with_leg=True, is_log=True, is_int_x=False):
    if with_leg:
        ax.legend(loc='best', frameon=True)
    if is_log:
        ax.semilogy()
    if is_int_x:
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.grid(ls=":")
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()



import numpy as np
import teneva


from ..equation import Equation


class EquationDum(Equation):
    """Dumbbell model, a special case of the FPE (it works only for d=3)."""

    def __init__(self, d=3, e=1.E-5, is_full=False, name='Dum'):
        super().__init__(d, e, is_full, name)

        if is_full:
            raise ValueError('This equation works only for TT-format')

        self.set_grid(n=60, a=-10., b=+10.)
        self.set_grid_time(m=100, t=10.)

        self.eta_list = []
        self.psi_list = []

        self.eta_real = +1.0328125
        self.psi_real = +2.071143

    def calc_eta(self, A):
        def func(X):
            R = teneva.cheb_get(X, A, self.a, self.b)
            return self.deta(X, R)

        Y_ = self.build(func)
        A_ = teneva.cheb_int(Y_)

        eta = teneva.cheb_sum(A_, self.a, self.b)
        self.eta_list.append(eta)

    def calc_psi(self, A):
        def func(X):
            R = teneva.cheb_get(X, A, self.a, self.b)
            return self.dpsi(X, R)

        Y_ = self.build(func)
        A_ = teneva.cheb_int(Y_)

        psi = teneva.cheb_sum(A_, self.a, self.b)
        self.psi_list.append(psi)

    def callback(self, solver):
        if not solver.is_last:
            return

        self.calc_eta(solver.A)
        self.calc_psi(solver.A)

        e_eta = abs(self.eta_real - self.eta_list[-1]) / abs(self.eta_real)
        e_psi = abs(self.psi_real - self.psi_list[-1]) / abs(self.psi_real)

        return f' | e_eta={e_eta:-8.2e} | e_psi={e_psi:-8.2e}'

    def dphi(self, X):
        X = X.T
        res = np.exp(-np.sum(X*X, axis=0) / 2. / self.p**2)
        res = 1. - self.a_opt / self.p**5 * res
        res = res * X
        return res.T

    def deta(self, X, r):
        p = self.dphi(X)
        return r * X[:, 0] * p[:, 1]  / self.b_opt

    def dpsi(self, X, r):
        p = self.dphi(X)
        return r * (X[:, 0] * p[:, 0] - X[:, 1] * p[:, 1]) / self.b_opt**2

    def f(self, X, t):
        X = X.T
        q = self.a_opt / 2. / self.p**5
        q *= np.exp(-np.sum(X*X, axis=0) / 2. / self.p**2)
        res = (q - 0.5) * X + self.K @ X
        return res.T

    def f1(self, X, t):
        X = X.T
        q = self.a_opt / 2. / self.p**5
        q *= np.exp(-np.sum(X*X, axis=0) / 2. / self.p**2)
        res = -0.5 + q - q / self.p**2 * X**2
        return res.T

    def init(self):
        self.with_rs = False
        self.with_rt = False

        self.a_opt = 0.1
        self.b_opt = 1.
        self.p = 0.5

        self.K = np.zeros((self.d, self.d))
        self.K[0, 1] = 1.
        self.K *= self.b_opt



"""Demo script that shows how to use the Fokker-Planck solver.

Run it from the root of the project as "clear && python demo/demo.py". The
output should look like this (see also the plot "demo/demo_result.png"):
"
Solve: 100step [00:06, 15.92step/s,  | t=2.50 | r=1.0 | e_t=1.51e-02]

The value of PDF at poi=[0, 0, 0] for different time moments:
Time   0.0250 | Value   0.0433
Time   0.0500 | Value   0.0422
Time   0.0750 | Value   0.0413
...
Time   2.4750 | Value   0.0095
Time   2.5000 | Value   0.0094
".

Note:
    To create the new equation, the base (abstract) class "Equation" should be
    inherited (see the class "EquationOwn" below). The detailed description of
    the base class methods and the format of input and output arguments is
    presented in the module "fpcross/equation.py" (online documentation will be
    available soon).

    Numerous parameters of the TT-cross method can be refined when calling the
    "Equation" class method "set_cross_opts". See the documentation for this
    function for more details. At the same time, for all prepared model
    examples (multidimensional simple diffusion problem, multidimensional
    Ornstein-Uhlenbeck process and 3-dimensional dumbbell model), a high quality
    solution is obtained with the default values of these parameters.

    For the methods "f", "f1", "r0", "rs", "rt" of the "Equation" class, input
    variable "X" is np.ndarray of the shape [samples, dimensions]. The output
    should be np.ndarray of the shape [samples, dimensions] for "f" and "f1",
    and np.ndarray of the shape [samples] for "r0", "rs" and "rt".

"""
import numpy as np


from fpcross import Equation
from fpcross import FPCross


class EquationOwn(Equation):
    """Fokker-Planck equation, which relates to the simple diffusion problem."""

    def callback(self, solver):
        """Function is called by solver after each time step."""

        # We calculate the value of the solution at some point at each
        # time step (just for demonstration purposes):
        self.t_list.append(solver.t)
        self.y_list.append(solver.get(self.poi))

    def f(self, X, t):
        """The rhs."""
        return np.zeros(X.shape)

    def f1(self, X, t):
        """The diagonal derivatives for the rhs."""
        return np.zeros(X.shape)

    def init(self):
        """Initialize the equation parameters."""
        self.with_rs = False # Should be True, if "rs" function is set
        self.with_rt = True  # Should be True, if "rt" function is set

        # We will calculate the value of the solution at this point at each
        # time step (just for demonstration purposes):
        self.poi = np.zeros(self.d)
        self.t_list = []
        self.y_list = []

    def r0(self, X):
        """Initial PDF r(x, 0). The Gaussian distribution is used."""
        a = 2. * self.coef_pdf
        r = np.exp(-np.sum(X*X, axis=1) / a)
        r /= (np.pi * a)**(self.d/2)
        return r

    def rs(self, X):
        """Stationary PDF r0(x) for accuracy check."""
        return

    def rt(self, X, t):
        """Time-dependent real PDF r(x, t) for accuracy check."""
        a = 2. * self.coef_pdf + 4. * self.coef * t
        r = np.exp(-np.sum(X*X, axis=1) / a)
        r /= (np.pi * a)**(self.d/2)
        return r

    def set_coef_pdf(self, coef_pdf=1.):
        """Set coefficient for initial PDF Gaussian distribution.

        This function is presented in the base class "Equation" and it is called
        automatically by class constructor. We present it here for
        demonstration purposes only. Note that the value of this parameter is
        used in both "r0" and "rt" methods.

        """
        self.coef_pdf = coef_pdf

# First we have to create a description of the equation to be solved:
eq = EquationOwn(
    d=3,                          # Number of dimensions.
    e=1.E-6,                      # Desired accuracy of approximation (for TT)
    is_full=False,                # Use full (numpy) format or TT-format
    name='My equation')           # Display name
eq.set_grid(n=20, a=-10., b=+10.) # Number of spatial points, lower/upper bounds
eq.set_grid_time(m=100, t=2.5)    # Number of time points, upper bound for time
eq.init()                         # Call it after setting all equation params

# Now we can solve the eqation:
fpc = FPCross(
    eq,
    with_hist=True) # Accuracy will be checked after each step with this flag
fpc.solve()
fpc.plot('./demo/demo_result.png')

# We may print the computed values in selected point from each time step:
print('\nThe value of PDF at poi=[0, 0, 0] for different time moments:')
for t, y in zip(eq.t_list, eq.y_list):
    print(f'Time {t:-8.4f} | Value {y:-8.4f}')



import numpy as np
from scipy.linalg import expm
from scipy.linalg import toeplitz
import teneva
from time import perf_counter as tpc
from tqdm import tqdm


from .plot import plot
from .plot import plot_spec


class FPCross:
    def __init__(self, eq, with_hist=False, with_y_list=False, with_a_list=False):
        """Class that represents the solver for the Fokker-Planck equation.

        Args:
            eq (Equation): equation class instance.
            with_hist (bool): if flag is set, then accuracy of the result will
                be checked after each time step. Otherwise, the accuracy check
                will be performed only after the solver has finished running.
            with_y_list (bool): if flag is set, then solutions from each time
                step will be saved in "Y_list" variable.
            with_a_list (bool): if flag is set, then interpoltation
                coefficients of the solutions from each time step will be saved
                in "A_list" variable.

        """
        self.eq = eq
        self.is_full = self.eq.is_full
        self.with_hist = bool(with_hist)

        self.A = None  # Interpolation coefficients for the Y
        self.W = None  # Copy of the conv. solution from the previous time step
        self.Y = None  # Current solution r(x, t); it is tensor on the grid
        self.Z = None  # Matrix exponent for the finite difference matrices
        self.m = 0     # Current time-step
        self.s = None  # Current value of Y integral over the spatial domain
        self.t = 0.    # Current value of the time variable
        self.tc = 0.   # Computation time (duration of solver work)

        # Special varuables:
        self.is_last = False

        # History for errors, ranks ant integral while computation proccess:
        self.es_list = []
        self.et_list = []
        self.r_list = []
        self.s_list = []

        # History for solutions from each time step:
        self.with_y_list = bool(with_y_list)
        self.Y_list = []
        self.with_a_list = bool(with_a_list)
        self.A_list = []

    def get(self, X):
        """Calculate the current solution in the given spatial point.

        Args:
            X (np.ndarray): the spatial point (or several points) of interest.
                It is np.ndarray of the shape [dimensions] or [samples,
                dimensions].

        Returns:
            float or np.ndarray: the solution value for the given spatial
            point. If only one point is provided (i.e., X has the shape
            [dimensions]), then it will be float. If several points are
            provided (i.e., X has the shape [samples, dimensions]), then it
            will be np.ndarray of the shape [samples].

        """
        is_one = len(X.shape) == 1
        if is_one:
            X = X.reshape(1, -1)

        func = teneva.cheb_get_full if self.is_full else teneva.cheb_get
        y = func(X, self.A, self.eq.a, self.eq.b)

        return y[0] if is_one else y

    def plot(self, fpath=None, is_spec=False):
        """Plot the computation result.

        Args:
            fpath (str): optional path to png-file to save the figure.
            is_spec (bool): this flag should be set for the "EquationDum".

        """

        if is_spec:
            plot_spec(self, fpath)
        else:
            plot(self, fpath)

    def solve(self):
        """Solve the Fokker-Planck equation."""
        tqdm_ = tqdm(desc='Solve', unit='step', total=self.eq.m-1, ncols=90)
        self._step_init()

        for m in range(1, self.eq.m + 1):
            self.m = m
            self.t = self.eq.h * self.m
            self.is_last = (m == self.eq.m)

            self._step()
            self._step_proc()
            self.text += self.eq.callback(self) or ''

            tqdm_.set_postfix_str(self.text, refresh=True)
            tqdm_.update(1)

        tqdm_.close()

    def _check_rs(self):
        if not self.eq.with_rs:
            return

        if self.eq.Ys is None:
            self.eq.build_rs()

        e = teneva.accuracy(self.Y, self.eq.Ys)
        self.es_list.append(e)

        self.text += f' | e_s={e:-8.2e}'
        return e

    def _check_rt(self):
        if not self.eq.with_rt:
            return

        self.eq.build_rt(self.t)

        e = teneva.accuracy(self.Y, self.eq.Yt)
        self.et_list.append(e)

        self.text += f' | e_t={e:-8.2e}'
        return e

    def _conv_apply(self, n_fine_fact=None):
        if n_fine_fact is not None and not self.is_full:
            # Experimental option!! Should be "None" in actual code
            n_fine = self.eq.n * n_fine_fact
            Z = teneva.cheb_gets(self.A, self.eq.a, self.eq.b, n_fine)

        def func(y, t):
            X, r = y[:, :-1], y[:, -1]
            f0 = self.eq.f(X, self.t)
            f1 = self.eq.f1(X, self.t)
            f1_mult = - np.sum(f1, axis=1) * r
            f1_mult = f1_mult.reshape(-1, 1)
            return np.hstack([f0, f1_mult])

        def step_conv(X):
            X0 = ode_solve(self.eq.f, X, self.t, 2, -self.eq.h)

            if self.is_full:
                w0 = teneva.cheb_get_full(X0, self.A, self.eq.a, self.eq.b)
            else:
                if n_fine_fact is None:
                    w0 = teneva.cheb_get(X0, self.A, self.eq.a, self.eq.b)
                else:
                    I0 = teneva.poi_to_ind(X0,
                        self.eq.a, self.eq.b, n_fine, 'cheb')
                    w0 = teneva.get_many(Z, I0)

            y0 = np.hstack([X0, w0.reshape(-1, 1)])

            y1 = ode_solve(func, y0, self.t-self.eq.h, 2, self.eq.h)

            w1 = y1[:, -1]

            return w1

        self.Y = self.eq.build(step_conv, self.W)

    def _diff_apply(self):
        if self.is_full:
            self.Y = self.Y.reshape(-1, order='F')
            self.Y = self.Z @ self.Y
            self.Y = self.Y.reshape(self.eq.n, order='F')
        else:
            for k in range(self.eq.d):
                self.Y[k] = np.einsum('ij,kjm->kim', self.Z, self.Y[k])
            # TODO. Do we need truncation here (?):
            self.Y = teneva.truncate(self.Y, self.eq.e)

    def _diff_init(self):
        n = self.eq.n[0]
        D1, D2 = teneva.cheb_diff_matrix(self.eq.a[0], self.eq.b[0], n, 2)
        h0 = self.eq.h / 2
        J0 = np.eye(n)
        J0[+0, +0] = 0.
        J0[-1, -1] = 0.
        self.Z = expm(h0 * self.eq.coef * J0 @ D2)

        if self.is_full:
            Z0 = self.Z.copy()
            for k in range(1, self.eq.d):
                self.Z = np.kron(self.Z, Z0)

    def _interpolate(self):
        if self.is_full:
            self.A = teneva.cheb_int_full(self.Y)
        else:
            self.A = teneva.cheb_int(self.Y)
            # TODO. Do we need truncation here (?):
            self.A = teneva.truncate(self.A, self.eq.e)

    def _renormalize(self):
        if self.is_full:
            self.s = teneva.cheb_sum_full(self.A, self.eq.a, self.eq.b)
            self.Y = (1. / self.s) * self.Y
        else:
            self.s = teneva.cheb_sum(self.A, self.eq.a, self.eq.b)
            self.Y = teneva.mul(1. / self.s, self.Y)

    def _step(self):
        tc = tpc()
        self._diff_apply()
        self._interpolate()
        self._conv_apply()
        #self._interpolate()
        #self._renormalize()
        self.W = teneva.copy(self.Y)
        self._diff_apply()
        self._interpolate()
        self.tc += tpc() - tc

        self._step_hist()

    def _step_hist(self):
        if self.with_y_list:
            Y = self.Y.copy() if self.is_full else teneva.copy(self.Y)
            self.Y_list.append(Y)
        if self.with_a_list:
            A = self.A.copy() if self.is_full else teneva.copy(self.A)
            self.A_list.append(A)

    def _step_init(self):
        tc = tpc()
        self._diff_init()
        self.Y = self.eq.build_r0()

        self._interpolate()
        self._renormalize()



        self.W = teneva.copy(self.Y)
        self.tc += tpc() - tc

    def _step_proc(self):
        self.text = f' | t={self.t:-4.2f}'

        if not self.is_full:
            r = teneva.erank(self.Y)
            self.r_list.append(r)
            self.text += f' | r={r:-3.1f}'

        if self.with_hist or self.is_last:
            self._check_rs()
            self._check_rt()

        self.s_list.append(self.s)


def ode_solve(f, y0, t, n, h):
    y = y0.copy()
    for _ in range(1, n):
        k1 = h * f(y, t)
        k2 = h * f(y + 0.5 * k1, t + 0.5 * h)
        k3 = h * f(y + 0.5 * k2, t + 0.5 * h)
        k4 = h * f(y + k3, t + h)
        y += (k1 + 2 * (k2 + k3) + k4) / 6.
        t += h
    return y



import numpy as np


from ..equation import Equation


class EquationDif(Equation):
    """Simple diffusion problem, a special case of the FPE."""

    def __init__(self, d=3, e=1.E-6, is_full=False, name='Dif'):
        super().__init__(d, e, is_full, name)

        self.set_grid(n=20, a=-10., b=+10.)
        self.set_grid_time(m=100, t=2.5)

    def f(self, X, t):
        return np.zeros(X.shape)

    def f1(self, X, t):
        return np.zeros(X.shape)

    def init(self):
        self.with_rs = False
        self.with_rt = True

    def rt(self, X, t):
        a = 2. * self.coef_pdf + 4. * self.coef * t
        r = np.exp(-np.sum(X*X, axis=1) / a)
        r /= (np.pi * a)**(self.d/2)
        return r



import numpy as np


from fpcross import EquationOUP
from fpcross import EquationOUPUniv
from fpcross import FPCross
from fpcross import FPCrossOUPDiag


def run(d, A, m, t, is_new=True):
    Equation_ = EquationOUPUniv if is_new else EquationOUP
    FPCross_ = FPCrossOUPDiag if is_new else FPCross

    eq = Equation_(d)
    eq.set_grid_time(m, t)
    eq.set_coef_rhs(A)
    eq.init()

    fpc = FPCross_(eq, with_hist=True)
    fpc.solve()
    fpc.plot('./demo/result/oup_3d_test_' + ('new' if is_new else 'old'))


if __name__ == '__main__':
    a = [1.5, 1.0, 0.5, 2.0, 1.5, 1.2, 2.2, 3.3, 1.1, 5.1]
    A = np.diag(a)  # RHS matrix
    d = A.shape[0]  # Dimension
    m = 200         # Number of time points
    t = 10.         # Final time

    run(d, A, m, t)                # Run new solver (for diag OUP)
    # run(d, A, m, t, is_new=False)  # Run old solver (for general FPE)

    # Run it as "python demo/demo_oup_spec.py"
